use std::time::{Duration, Instant};

use aoc::util::parse::ParseExt;

struct Solution {
    year: u32,
    day: u32,
    input: &'static str,
    run: fn(&str) -> (String, Duration, String, Duration),
}

macro_rules! solution {
    ($year:tt, $day:tt) => {
        Solution {
            year: (&stringify!($year)).parse_u32(),
            day: (&stringify!($day)).parse_u32(),
            input: include_str!(concat![
                "../input/",
                stringify!($year),
                "/",
                stringify!($day),
                ".txt"
            ]),
            run: |input: &str| {
                use aoc::$year::$day::*;
                let part1_start = Instant::now();
                let part1 = part1(&input).to_string();
                let part1_elapsed = part1_start.elapsed();
                let part2_start = Instant::now();
                let part2 = part2(&input).to_string();
                let part2_elapsed = part2_start.elapsed();
                (part1, part1_elapsed, part2, part2_elapsed)
            },
        }
    };
}

fn main() {
    let (year, day) = match std::env::args().nth(1) {
        Some(arg) => {
            let mut split = arg.split("::").map(|s| s.parse_u32());
            (split.next(), split.next())
        }
        None => (None, None),
    };
    let solutions: Vec<_> = solutions()
        .filter(|s| year == Some(s.year) || year.is_none())
        .filter(|s| day == Some(s.day) || day.is_none())
        .collect();

    let count = solutions.len();
    let mut overall_duration = Duration::new(0, 0);
    for Solution {
        year,
        day,
        input,
        run,
    } in solutions
    {
        let start = Instant::now();
        let (aw1, el1, aw2, el2) = (run)(input);
        let elapsed = start.elapsed();
        overall_duration += elapsed;
        println!("{year} Day {day:02} ({elapsed:?})");
        println!("    Part 1: {aw1} ({el1:?})");
        println!("    Part 2: {aw2} ({el2:?})");
    }
    println!("Total solutions: {count} ({overall_duration:?})")
}

fn solutions() -> impl Iterator<Item = Solution> {
    std::iter::empty()
        .chain(year2016())
        .chain(year2022())
        .chain(year2023())
}

fn year2016() -> Vec<Solution> {
    vec![
        solution!(year2016, day01),
        solution!(year2016, day02),
        solution!(year2016, day03),
    ]
}

fn year2022() -> Vec<Solution> {
    vec![
        solution!(year2022, day01),
        solution!(year2022, day02),
        solution!(year2022, day03),
        solution!(year2022, day04),
        solution!(year2022, day05),
    ]
}

fn year2023() -> Vec<Solution> {
    vec![solution!(year2023, day01)]
}
